@@ -1,9 +1,9 @@
 /*
  * security/ccsecurity/internal.h
  *
- * Copyright (C) 2005-2011  NTT DATA CORPORATION
+ * Copyright (C) 2005-2012  NTT DATA CORPORATION
  *
- * Version: 1.8.0+   2011/02/07
+ * Version: 1.8.3+   2011/11/11
  */
 
 #ifndef _SECURITY_CCSECURITY_INTERNAL_H
@@ -16,7 +16,9 @@
 #include <linux/mm.h>
 #include <linux/utime.h>
 #include <linux/file.h>
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 38)
 #include <linux/smp_lock.h>
+#endif
 #include <linux/module.h>
 #include <linux/init.h>
 #include <linux/slab.h>
@@ -78,6 +80,12 @@ struct path {
 
 #endif
 
+#ifndef __printf
+#define __printf(a,b) __attribute__((format(printf,a,b)))
+#endif
+#ifndef __packed
+#define __packed __attribute__((__packed__))
+#endif
 #ifndef bool
 #define bool _Bool
 #endif
@@ -268,6 +276,44 @@ static inline void list_add_rcu(struct list_head *new, struct list_head *head)
 
 #endif
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 38)
+
+/**
+ * __list_del_entry - Deletes entry from list without re-initialization.
+ *
+ * @entry: Pointer to "struct list_head".
+ *
+ * Returns nothing.
+ *
+ * This is for compatibility with older kernels.
+ */
+static inline void __list_del_entry(struct list_head *entry)
+{
+	__list_del(entry->prev, entry->next);
+}
+
+#endif
+
+#ifndef list_for_each_entry_safe
+
+/**
+ * list_for_each_entry_safe - Iterate over list of given type safe against removal of list entry.
+ *
+ * @pos:    The "type *" to use as a loop cursor.
+ * @n:      Another "type *" to use as temporary storage.
+ * @head:   Pointer to "struct list_head".
+ * @member: The name of the list_struct within the struct.
+ *
+ * This is for compatibility with older kernels.
+ */
+#define list_for_each_entry_safe(pos, n, head, member)                  \
+	for (pos = list_entry((head)->next, typeof(*pos), member),      \
+		     n = list_entry(pos->member.next, typeof(*pos), member); \
+	     &pos->member != (head);					\
+	     pos = n, n = list_entry(n->member.next, typeof(*n), member))
+
+#endif
+
 #ifndef srcu_dereference
 
 /**
@@ -292,7 +338,7 @@ static inline void list_add_rcu(struct list_head *new, struct list_head *head)
  * @pos:    The type * to use as a loop cursor.
  * @head:   The head for your list.
  * @member: The name of the list_struct within the struct.
- * @ss:     Pointer to "struct srcu_struct".  
+ * @ss:     Pointer to "struct srcu_struct".
  *
  * As of 2.6.36, this macro is not provided because only TOMOYO wants it.
  */
@@ -334,17 +380,46 @@ static inline void list_add_rcu(struct list_head *new, struct list_head *head)
  * TOMOYO specific part start.
  */
 
+/* Clear TOMOYO 1.8 config. */
+#undef CONFIG_CCSECURITY
+#undef CONFIG_CCSECURITY_LKM
+#undef CONFIG_CCSECURITY_DISABLE_BY_DEFAULT
+#undef CONFIG_CCSECURITY_MAX_AUDIT_LOG
+#undef CONFIG_CCSECURITY_MAX_ACCEPT_ENTRY
+#undef CONFIG_CCSECURITY_OMIT_USERSPACE_LOADER
+#undef CONFIG_CCSECURITY_POLICY_LOADER
+#undef CONFIG_CCSECURITY_ACTIVATION_TRIGGER
+#undef CONFIG_CCSECURITY_USE_EXTERNAL_TASK_SECURITY
+#undef CONFIG_CCSECURITY_FILE_READDIR
+#undef CONFIG_CCSECURITY_FILE_GETATTR
+#undef CONFIG_CCSECURITY_NETWORK
+#undef CONFIG_CCSECURITY_NETWORK_RECVMSG
+#undef CONFIG_CCSECURITY_CAPABILITY
+#undef CONFIG_CCSECURITY_IPC
+#undef CONFIG_CCSECURITY_MISC
+#undef CONFIG_CCSECURITY_TASK_EXECUTE_HANDLER
+#undef CONFIG_CCSECURITY_TASK_DOMAIN_TRANSITION
+#undef CONFIG_CCSECURITY_PORTRESERVE
+/* Define AKARI 1.0 config. */
+#include "config.h"
 #ifndef CONFIG_CCSECURITY
 #define CONFIG_CCSECURITY
-#define CONFIG_CCSECURITY_MAX_AUDIT_LOG    1024
-#define CONFIG_CCSECURITY_MAX_ACCEPT_ENTRY 2048
-#define CONFIG_CCSECURITY_DEFAULT_LOADER "/sbin/ccs-init"
-#define CONFIG_CCSECURITY_ALTERNATIVE_TRIGGER "/sbin/ccs-start"
-#define CONFIG_CCSECURITY_BUILTIN_INITIALIZERS ""
 #endif
 #ifndef CONFIG_CCSECURITY_USE_EXTERNAL_TASK_SECURITY
 #define CONFIG_CCSECURITY_USE_EXTERNAL_TASK_SECURITY
 #endif
+#ifndef CONFIG_CCSECURITY_MAX_AUDIT_LOG
+#define CONFIG_CCSECURITY_MAX_AUDIT_LOG 1024
+#endif
+#ifndef CONFIG_CCSECURITY_MAX_ACCEPT_ENTRY
+#define CONFIG_CCSECURITY_MAX_ACCEPT_ENTRY 2048
+#endif
+#ifndef CONFIG_CCSECURITY_POLICY_LOADER
+#define CONFIG_CCSECURITY_POLICY_LOADER "/sbin/ccs-init"
+#endif
+#ifndef CONFIG_CCSECURITY_ACTIVATION_TRIGGER
+#define CONFIG_CCSECURITY_ACTIVATION_TRIGGER "/sbin/init"
+#endif
 #include "ccsecurity.h"
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
 #error This module supports only 2.6.0 and later kernels.
@@ -371,15 +446,27 @@ enum ccs_acl_entry_type_index {
 	CCS_TYPE_PATH_NUMBER_ACL,
 	CCS_TYPE_MKDEV_ACL,
 	CCS_TYPE_MOUNT_ACL,
+#ifdef CONFIG_CCSECURITY_MISC
 	CCS_TYPE_ENV_ACL,
+#endif
+#ifdef CONFIG_CCSECURITY_CAPABILITY
 	CCS_TYPE_CAPABILITY_ACL,
+#endif
+#ifdef CONFIG_CCSECURITY_NETWORK
 	CCS_TYPE_INET_ACL,
 	CCS_TYPE_UNIX_ACL,
+#endif
+#ifdef CONFIG_CCSECURITY_IPC
 	CCS_TYPE_SIGNAL_ACL,
+#endif
+#ifdef CONFIG_CCSECURITY_TASK_EXECUTE_HANDLER
 	CCS_TYPE_AUTO_EXECUTE_HANDLER,
 	CCS_TYPE_DENIED_EXECUTE_HANDLER,
+#endif
+#ifdef CONFIG_CCSECURITY_TASK_DOMAIN_TRANSITION
 	CCS_TYPE_AUTO_TASK_ACL,
 	CCS_TYPE_MANUAL_TASK_ACL,
+#endif
 };
 
 /* Index numbers for "struct ccs_condition". */
@@ -481,24 +568,27 @@ enum ccs_grant_log {
 enum ccs_group_id {
 	CCS_PATH_GROUP,
 	CCS_NUMBER_GROUP,
+#ifdef CONFIG_CCSECURITY_NETWORK
 	CCS_ADDRESS_GROUP,
+#endif
 	CCS_MAX_GROUP
 };
 
-/* Index numbers for type of IP address. */
-enum ccs_ip_address_type {
-	CCS_IP_ADDRESS_TYPE_ADDRESS_GROUP,
-	CCS_IP_ADDRESS_TYPE_IPv4,
-	CCS_IP_ADDRESS_TYPE_IPv6,
-};
-
 /* Index numbers for category of functionality. */
 enum ccs_mac_category_index {
 	CCS_MAC_CATEGORY_FILE,
+#ifdef CONFIG_CCSECURITY_NETWORK
 	CCS_MAC_CATEGORY_NETWORK,
+#endif
+#ifdef CONFIG_CCSECURITY_MISC
 	CCS_MAC_CATEGORY_MISC,
+#endif
+#ifdef CONFIG_CCSECURITY_IPC
 	CCS_MAC_CATEGORY_IPC,
+#endif
+#ifdef CONFIG_CCSECURITY_CAPABILITY
 	CCS_MAC_CATEGORY_CAPABILITY,
+#endif
 	CCS_MAX_MAC_CATEGORY_INDEX
 };
 
@@ -508,7 +598,9 @@ enum ccs_mac_index {
 	CCS_MAC_FILE_OPEN,
 	CCS_MAC_FILE_CREATE,
 	CCS_MAC_FILE_UNLINK,
+#ifdef CONFIG_CCSECURITY_FILE_GETATTR
 	CCS_MAC_FILE_GETATTR,
+#endif
 	CCS_MAC_FILE_MKDIR,
 	CCS_MAC_FILE_RMDIR,
 	CCS_MAC_FILE_MKFIFO,
@@ -527,29 +619,42 @@ enum ccs_mac_index {
 	CCS_MAC_FILE_MOUNT,
 	CCS_MAC_FILE_UMOUNT,
 	CCS_MAC_FILE_PIVOT_ROOT,
+#ifdef CONFIG_CCSECURITY_NETWORK
 	CCS_MAC_NETWORK_INET_STREAM_BIND,
 	CCS_MAC_NETWORK_INET_STREAM_LISTEN,
 	CCS_MAC_NETWORK_INET_STREAM_CONNECT,
 	CCS_MAC_NETWORK_INET_STREAM_ACCEPT,
 	CCS_MAC_NETWORK_INET_DGRAM_BIND,
 	CCS_MAC_NETWORK_INET_DGRAM_SEND,
+#ifdef CONFIG_CCSECURITY_NETWORK_RECVMSG
 	CCS_MAC_NETWORK_INET_DGRAM_RECV,
+#endif
 	CCS_MAC_NETWORK_INET_RAW_BIND,
 	CCS_MAC_NETWORK_INET_RAW_SEND,
+#ifdef CONFIG_CCSECURITY_NETWORK_RECVMSG
 	CCS_MAC_NETWORK_INET_RAW_RECV,
+#endif
 	CCS_MAC_NETWORK_UNIX_STREAM_BIND,
 	CCS_MAC_NETWORK_UNIX_STREAM_LISTEN,
 	CCS_MAC_NETWORK_UNIX_STREAM_CONNECT,
 	CCS_MAC_NETWORK_UNIX_STREAM_ACCEPT,
 	CCS_MAC_NETWORK_UNIX_DGRAM_BIND,
 	CCS_MAC_NETWORK_UNIX_DGRAM_SEND,
+#ifdef CONFIG_CCSECURITY_NETWORK_RECVMSG
 	CCS_MAC_NETWORK_UNIX_DGRAM_RECV,
+#endif
 	CCS_MAC_NETWORK_UNIX_SEQPACKET_BIND,
 	CCS_MAC_NETWORK_UNIX_SEQPACKET_LISTEN,
 	CCS_MAC_NETWORK_UNIX_SEQPACKET_CONNECT,
 	CCS_MAC_NETWORK_UNIX_SEQPACKET_ACCEPT,
+#endif
+#ifdef CONFIG_CCSECURITY_MISC
 	CCS_MAC_ENVIRON,
+#endif
+#ifdef CONFIG_CCSECURITY_IPC
 	CCS_MAC_SIGNAL,
+#endif
+#ifdef CONFIG_CCSECURITY_CAPABILITY
 	CCS_MAC_CAPABILITY_USE_ROUTE_SOCKET,
 	CCS_MAC_CAPABILITY_USE_PACKET_SOCKET,
 	CCS_MAC_CAPABILITY_SYS_REBOOT,
@@ -560,10 +665,11 @@ enum ccs_mac_index {
 	CCS_MAC_CAPABILITY_USE_KERNEL_MODULE,
 	CCS_MAC_CAPABILITY_SYS_KEXEC_LOAD,
 	CCS_MAC_CAPABILITY_SYS_PTRACE,
+#endif
 	CCS_MAX_MAC_INDEX
 };
 
-/* Index numbers for /proc/ccs/meminfo interface. */
+/* Index numbers for /proc/ccs/stat interface. */
 enum ccs_memory_stat_type {
 	CCS_MEMORY_POLICY,
 	CCS_MEMORY_AUDIT,
@@ -597,7 +703,9 @@ enum ccs_network_acl_index {
 	CCS_NETWORK_CONNECT, /* connect() operation. */
 	CCS_NETWORK_ACCEPT,  /* accept() operation. */
 	CCS_NETWORK_SEND,    /* send() operation. */
+#ifdef CONFIG_CCSECURITY_NETWORK_RECVMSG
 	CCS_NETWORK_RECV,    /* recv() operation. */
+#endif
 	CCS_MAX_NETWORK_OPERATION
 };
 
@@ -616,7 +724,9 @@ enum ccs_path_acl_index {
 	CCS_TYPE_WRITE,
 	CCS_TYPE_APPEND,
 	CCS_TYPE_UNLINK,
+#ifdef CONFIG_CCSECURITY_FILE_GETATTR
 	CCS_TYPE_GETATTR,
+#endif
 	CCS_TYPE_RMDIR,
 	CCS_TYPE_TRUNCATE,
 	CCS_TYPE_SYMLINK,
@@ -650,15 +760,18 @@ enum ccs_path_stat_index {
 
 /* Index numbers for entry type. */
 enum ccs_policy_id {
+#ifdef CONFIG_CCSECURITY_PORTRESERVE
 	CCS_ID_RESERVEDPORT,
+#endif
 	CCS_ID_GROUP,
+#ifdef CONFIG_CCSECURITY_NETWORK
 	CCS_ID_ADDRESS_GROUP,
+#endif
 	CCS_ID_PATH_GROUP,
 	CCS_ID_NUMBER_GROUP,
 	CCS_ID_AGGREGATOR,
 	CCS_ID_TRANSITION_CONTROL,
 	CCS_ID_MANAGER,
-	CCS_ID_IPV6_ADDRESS,
 	CCS_ID_CONDITION,
 	CCS_ID_NAME,
 	CCS_ID_ACL,
@@ -686,25 +799,18 @@ enum ccs_pref_index {
 
 /* Index numbers for /proc/ccs/ interfaces. */
 enum ccs_proc_interface_index {
-	CCS_DOMAINPOLICY,
-	CCS_EXCEPTIONPOLICY,
-	CCS_DOMAIN_STATUS,
+	CCS_DOMAIN_POLICY,
+	CCS_EXCEPTION_POLICY,
 	CCS_PROCESS_STATUS,
-	CCS_MEMINFO,
 	CCS_STAT,
 	CCS_AUDIT,
 	CCS_VERSION,
 	CCS_PROFILE,
 	CCS_QUERY,
 	CCS_MANAGER,
+#ifdef CONFIG_CCSECURITY_TASK_EXECUTE_HANDLER
 	CCS_EXECUTE_HANDLER,
-};
-
-/* Index numbers for shared entries. */
-enum ccs_shared_acl_id {
-	CCS_CONDITION_LIST,
-	CCS_IPV6ADDRESS_LIST,
-	CCS_MAX_LIST
+#endif
 };
 
 /* Index numbers for special mount operations. */
@@ -722,6 +828,8 @@ enum ccs_special_mount {
 /* Index numbers for domain transition control keywords. */
 enum ccs_transition_type {
 	/* Do not change this order, */
+	CCS_TRANSITION_CONTROL_NO_RESET,
+	CCS_TRANSITION_CONTROL_RESET,
 	CCS_TRANSITION_CONTROL_NO_INITIALIZE,
 	CCS_TRANSITION_CONTROL_INITIALIZE,
 	CCS_TRANSITION_CONTROL_NO_KEEP,
@@ -753,13 +861,12 @@ enum ccs_value_type {
  */
 #define CCS_SOCK_MAX 6
 
-/* A domain definition starts with <kernel>. */
-#define CCS_ROOT_NAME                         "<kernel>"
-#define CCS_ROOT_NAME_LEN                     (sizeof(CCS_ROOT_NAME) - 1)
-
 /* Size of temporary buffer for execve() operation. */
 #define CCS_EXEC_TMPSIZE     4096
 
+/* Garbage collector is trying to kfree() this element. */
+#define CCS_GC_IN_PROGRESS -1
+
 /* Profile number is an integer between 0 and 255. */
 #define CCS_MAX_PROFILES 256
 
@@ -805,37 +912,36 @@ enum ccs_value_type {
 		       __GFP_NORETRY | __GFP_NOMEMALLOC)
 
 /* Size of read buffer for /proc/ccs/ interface. */
-#define CCS_MAX_IO_READ_QUEUE 32
+#define CCS_MAX_IO_READ_QUEUE 64
 
 /* Structure definition for internal use. */
 
 /* Common header for holding ACL entries. */
 struct ccs_acl_head {
 	struct list_head list;
-	bool is_deleted;
-} __attribute__((__packed__));
+	s8 is_deleted; /* true or false or CCS_GC_IN_PROGRESS */
+} __packed;
 
 /* Common header for shared entries. */
 struct ccs_shared_acl_head {
 	struct list_head list;
 	atomic_t users;
-} __attribute__((__packed__));
+} __packed;
 
 /* Common header for individual entries. */
 struct ccs_acl_info {
 	struct list_head list;
 	struct ccs_condition *cond; /* Maybe NULL. */
-	bool is_deleted;
-	u8 type; /* = one of values in "enum ccs_acl_entry_type_index" */
-} __attribute__((__packed__));
+	s8 is_deleted; /* true or false or CCS_GC_IN_PROGRESS */
+	u8 type; /* One of values in "enum ccs_acl_entry_type_index". */
+	u16 perm;
+} __packed;
 
 /* Structure for holding a word. */
 struct ccs_name_union {
 	/* Either @filename or @group is NULL. */
 	const struct ccs_path_info *filename;
 	struct ccs_group *group;
-	/* True if @group != NULL, false if @filename != NULL. */
-	u8 is_group;
 };
 
 /* Structure for holding a number. */
@@ -844,8 +950,13 @@ struct ccs_number_union {
 	struct ccs_group *group; /* Maybe NULL. */
 	/* One of values in "enum ccs_value_type". */
 	u8 value_type[2];
-	/* True if @group != NULL, false otherwise. */
-	u8 is_group;
+};
+
+/* Structure for holding an IP address. */
+struct ccs_ipaddr_union {
+	struct in6_addr ip[2]; /* Big endian. */
+	struct ccs_group *group; /* Pointer to address group. */
+	bool is_ipv6; /* Valid only if @group == NULL. */
 };
 
 /* Structure for "path_group"/"number_group"/"address_group" directive. */
@@ -875,11 +986,8 @@ struct ccs_number_group {
 /* Structure for "address_group" directive. */
 struct ccs_address_group {
 	struct ccs_acl_head head;
-	bool is_ipv6; /* True if IPv6 address, false if IPv4 address. */
-	union {
-		u32 ipv4;                    /* Host byte order    */
-		const struct in6_addr *ipv6; /* Network byte order */
-	} min, max;
+	/* Structure for holding an IP address. */
+	struct ccs_ipaddr_union address;
 };
 
 /* Subset of "struct stat". Used by conditional ACL and audit logs. */
@@ -887,7 +995,7 @@ struct ccs_mini_stat {
 	uid_t uid;
 	gid_t gid;
 	ino_t ino;
-	mode_t mode;
+	umode_t mode;
 	dev_t dev;
 	dev_t rdev;
 };
@@ -948,6 +1056,7 @@ struct ccs_condition {
 	u16 argc; /* Number of "struct ccs_argv". */
 	u16 envc; /* Number of "struct ccs_envp". */
 	u8 grant_log; /* One of values in "enum ccs_grant_log". */
+	bool exec_transit; /* True if transit is for "file execute". */
 	const struct ccs_path_info *transit; /* Maybe NULL. */
 	/*
 	 * struct ccs_condition_element condition[condc];
@@ -959,6 +1068,7 @@ struct ccs_condition {
 };
 
 struct ccs_execve;
+struct ccs_policy_namespace;
 
 /* Structure for request info. */
 struct ccs_request_info {
@@ -1013,10 +1123,10 @@ struct ccs_request_info {
 			 */
 			u8 operation;
 		} path_number;
+#ifdef CONFIG_CCSECURITY_NETWORK
 		struct {
-			const u32 *address;
-			u32 ip;
-			u16 port;
+			const u32 *address; /* Big endian. */
+			u16 port; /* Host endian. */
 			/* One of values smaller than CCS_SOCK_MAX. */
 			u8 protocol;
 			/* One of values in "enum ccs_network_acl_index". */
@@ -1030,17 +1140,24 @@ struct ccs_request_info {
 			/* One of values in "enum ccs_network_acl_index". */
 			u8 operation;
 		} unix_network;
+#endif
+#ifdef CONFIG_CCSECURITY_MISC
 		struct {
 			const struct ccs_path_info *name;
 		} environ;
+#endif
+#ifdef CONFIG_CCSECURITY_CAPABILITY
 		struct {
 			/* One of values in "enum ccs_capability_acl_index". */
 			u8 operation;
 		} capability;
+#endif
+#ifdef CONFIG_CCSECURITY_IPC
 		struct {
 			const char *dest_pattern;
 			int sig;
 		} signal;
+#endif
 		struct {
 			const struct ccs_path_info *type;
 			const struct ccs_path_info *dir;
@@ -1048,14 +1165,12 @@ struct ccs_request_info {
 			unsigned long flags;
 			int need_dev;
 		} mount;
+#ifdef CONFIG_CCSECURITY_TASK_DOMAIN_TRANSITION
 		struct {
 			const struct ccs_path_info *domainname;
 		} task;
+#endif
 	} param;
-	u8 param_type; /* One of values in "enum ccs_acl_entry_type_index". */
-	bool granted; /* True if granted, false otherwise. */
-	/* True if current thread should not be carried sleep penalty. */
-	bool dont_sleep_on_enforce_error;
 	/*
 	 * For updating current->ccs_domain_info at ccs_update_task_domain().
 	 * Initialized to NULL at ccs_init_request_info().
@@ -1063,6 +1178,10 @@ struct ccs_request_info {
 	 * granted. Re-initialized to NULL at ccs_update_task_domain().
 	 */
 	struct ccs_acl_info *matched_acl;
+	u8 param_type; /* One of values in "enum ccs_acl_entry_type_index". */
+	bool granted; /* True if granted, false otherwise. */
+	/* True if current thread should not be carried sleep penalty. */
+	bool dont_sleep_on_enforce_error;
 	/*
 	 * For counting number of retries made for this request.
 	 * This counter is incremented whenever ccs_supervisor() returned
@@ -1104,7 +1223,7 @@ struct ccs_execve {
 	struct ccs_obj_info obj;
 	struct linux_binprm *bprm;
 	struct ccs_domain_info *previous_domain;
-	int reader_idx;
+	const struct ccs_path_info *transition;
 	/* For execute_handler */
 	const struct ccs_path_info *handler;
 	char *handler_path; /* = kstrdup(handler->name, CCS_GFP_FLAGS) */
@@ -1117,25 +1236,27 @@ struct ccs_execve {
 /* Structure for domain information. */
 struct ccs_domain_info {
 	struct list_head list;
-	struct list_head acl_info_list[2];
+	struct list_head acl_info_list;
 	/* Name of this domain. Never NULL.          */
 	const struct ccs_path_info *domainname;
+	/* Namespace for this domain. Never NULL. */
+	struct ccs_policy_namespace *ns;
 	u8 profile;        /* Profile number to use. */
-	u8 group;
+	u8 group;          /* Group number to use.   */
 	bool is_deleted;   /* Delete flag.           */
 	bool flags[CCS_MAX_DOMAIN_INFO_FLAGS];
 };
 
 /*
- * Structure for "initialize_domain"/"no_initialize_domain" and
- * "keep_domain"/"no_keep_domain" keyword.
+ * Structure for "reset_domain"/"no_reset_domain"/"initialize_domain"/
+ * "no_initialize_domain"/"keep_domain"/"no_keep_domain" keyword.
  */
 struct ccs_transition_control {
 	struct ccs_acl_head head;
-	u8 type; /* = one of values in "enum ccs_transition_type" */
+	u8 type; /* One of values in "enum ccs_transition_type" */
 	bool is_last_name; /* True if the domainname is ccs_last_word(). */
-	const struct ccs_path_info *domainname;
-	const struct ccs_path_info *program;
+	const struct ccs_path_info *domainname; /* Maybe NULL */
+	const struct ccs_path_info *program;    /* Maybe NULL */
 };
 
 /* Structure for "aggregator" keyword. */
@@ -1148,8 +1269,7 @@ struct ccs_aggregator {
 /* Structure for "deny_autobind" keyword. */
 struct ccs_reserved {
 	struct ccs_acl_head head;
-	u16 min_port;                /* Start of port number range.          */
-	u16 max_port;                /* End of port number range.            */
+	struct ccs_number_union port;
 };
 
 /* Structure for policy manager. */
@@ -1162,7 +1282,7 @@ struct ccs_manager {
 
 /* Structure for argv[]. */
 struct ccs_argv {
-	unsigned int index;
+	unsigned long index;
 	const struct ccs_path_info *value;
 	bool is_not;
 };
@@ -1214,7 +1334,6 @@ struct ccs_task_acl {
  */
 struct ccs_path_acl {
 	struct ccs_acl_info head; /* type = CCS_TYPE_PATH_ACL */
-	u16 perm; /* Bitmask of values in "enum ccs_path_acl_index". */
 	struct ccs_name_union name;
 };
 
@@ -1223,7 +1342,6 @@ struct ccs_path_acl {
  */
 struct ccs_path2_acl {
 	struct ccs_acl_info head; /* type = CCS_TYPE_PATH2_ACL */
-	u8 perm; /* Bitmask of values in "enum ccs_path2_acl_index". */
 	struct ccs_name_union name1;
 	struct ccs_name_union name2;
 };
@@ -1234,7 +1352,6 @@ struct ccs_path2_acl {
  */
 struct ccs_path_number_acl {
 	struct ccs_acl_info head; /* type = CCS_TYPE_PATH_NUMBER_ACL */
-	u8 perm; /* Bitmask of values in "enum ccs_path_number_acl_index". */
 	struct ccs_name_union name;
 	struct ccs_number_union number;
 };
@@ -1242,7 +1359,6 @@ struct ccs_path_number_acl {
 /* Structure for "file mkblock" and "file mkchar" directive. */
 struct ccs_mkdev_acl {
 	struct ccs_acl_info head; /* type = CCS_TYPE_MKDEV_ACL */
-	u8 perm; /* Bitmask of values in "enum ccs_mkdev_acl_index". */
 	struct ccs_name_union name;
 	struct ccs_number_union mode;
 	struct ccs_number_union major;
@@ -1267,54 +1383,22 @@ struct ccs_env_acl {
 /* Structure for "capability" directive. */
 struct ccs_capability_acl {
 	struct ccs_acl_info head; /* type = CCS_TYPE_CAPABILITY_ACL */
-	u8 operation; /* one of values in "enum ccs_capability_acl_index". */
+	u8 operation; /* One of values in "enum ccs_capability_acl_index". */
 };
 
 /* Structure for "ipc signal" directive. */
 struct ccs_signal_acl {
 	struct ccs_acl_info head; /* type = CCS_TYPE_SIGNAL_ACL */
-	u16 sig;
+	struct ccs_number_union sig;
 	/* Pointer to destination pattern. */
 	const struct ccs_path_info *domainname;
 };
 
-/* Structure for holding an IPv6 address. */
-struct ccs_ipv6addr {
-	struct ccs_shared_acl_head head;
-	struct in6_addr addr;
-};
-
 /* Structure for "network inet" directive. */
 struct ccs_inet_acl {
 	struct ccs_acl_info head; /* type = CCS_TYPE_INET_ACL */
 	u8 protocol;
-	u8 perm; /* Bitmask of values in "enum ccs_network_acl_index" */
-	/*
-	 * address_type takes one of the following constants.
-	 *   CCS_IP_ADDRESS_TYPE_ADDRESS_GROUP
-	 *                if @address points to "address_group" directive.
-	 *   CCS_IP_ADDRESS_TYPE_IPv4
-	 *                if @address points to an IPv4 address.
-	 *   CCS_IP_ADDRESS_TYPE_IPv6
-	 *                if @address points to an IPv6 address.
-	 */
-	u8 address_type;
-	union {
-		struct {
-			/* Start of IPv4 address range. Host endian. */
-			u32 min;
-			/* End of IPv4 address range. Host endian.   */
-			u32 max;
-		} ipv4;
-		struct {
-			/* Start of IPv6 address range. Big endian.  */
-			const struct in6_addr *min;
-			/* End of IPv6 address range. Big endian.    */
-			const struct in6_addr *max;
-		} ipv6;
-		/* Pointer to address group. */
-		struct ccs_group *group;
-	} address;
+	struct ccs_ipaddr_union address;
 	struct ccs_number_union port;
 };
 
@@ -1322,7 +1406,6 @@ struct ccs_inet_acl {
 struct ccs_unix_acl {
 	struct ccs_acl_info head; /* type = CCS_TYPE_UNIX_ACL */
 	u8 protocol;
-	u8 perm; /* Bitmask of values in "enum ccs_network_acl_index" */
 	struct ccs_name_union name;
 };
 
@@ -1335,55 +1418,82 @@ struct ccs_name {
 
 /* Structure for holding a line from /proc/ccs/ interface. */
 struct ccs_acl_param {
-	char *data;
-	struct ccs_domain_info *domain;
-	bool is_delete;
+	char *data; /* Unprocessed data. */
+	struct list_head *list; /* List to add or remove. */
+	struct ccs_policy_namespace *ns; /* Namespace to use. */
+	bool is_delete; /* True if it is a delete request. */
+	union ccs_acl_union {
+		struct ccs_acl_info acl_info;
+		struct ccs_handler_acl handler_acl;
+		struct ccs_task_acl task_acl;
+		struct ccs_path_acl path_acl;
+		struct ccs_path2_acl path2_acl;
+		struct ccs_path_number_acl path_number_acl;
+		struct ccs_mkdev_acl mkdev_acl;
+		struct ccs_mount_acl mount_acl;
+		struct ccs_env_acl env_acl;
+		struct ccs_capability_acl capability_acl;
+		struct ccs_signal_acl signal_acl;
+		struct ccs_inet_acl inet_acl;
+		struct ccs_unix_acl unix_acl;
+		/**/
+		struct ccs_acl_head acl_head;
+		struct ccs_transition_control transition_control;
+		struct ccs_aggregator aggregator;
+		struct ccs_reserved reserved;
+		struct ccs_manager manager;
+		struct ccs_path_group path_group;
+		struct ccs_number_group number_group;
+		struct ccs_address_group address_group;
+	} e;
 };
 
-/* Structure for reading/writing policy via /proc interfaces. */
+/* Structure for reading/writing policy via /proc/ccs/ interfaces. */
 struct ccs_io_buffer {
-	void (*read) (struct ccs_io_buffer *);
-	int (*write) (struct ccs_io_buffer *);
-	int (*poll) (struct file *file, poll_table *wait);
 	/* Exclusive lock for this structure.   */
 	struct mutex io_sem;
-	/* Index returned by ccs_lock().        */
-	int reader_idx;
 	char __user *read_user_buf;
-	int read_user_buf_avail;
+	size_t read_user_buf_avail;
 	struct {
+		struct list_head *ns;
 		struct list_head *domain;
 		struct list_head *group;
 		struct list_head *acl;
-		int avail;
-		int step;
-		int query_index;
+		size_t avail;
+		unsigned int step;
+		unsigned int query_index;
 		u16 index;
 		u16 cond_index;
-		u8 group_index;
+		u8 acl_group_index;
 		u8 cond_step;
 		u8 bit;
 		u8 w_pos;
 		bool eof;
 		bool print_this_domain_only;
-		bool print_execute_only;
+		bool print_transition_related_only;
 		bool print_cond_part;
 		const char *w[CCS_MAX_IO_READ_QUEUE];
 	} r;
 	struct {
+		struct ccs_policy_namespace *ns;
 		struct ccs_domain_info *domain;
-		int avail;
+		size_t avail;
+		bool is_delete;
 	} w;
 	/* Buffer for reading.                  */
 	char *read_buf;
 	/* Size of read buffer.                 */
-	int readbuf_size;
+	size_t readbuf_size;
 	/* Buffer for writing.                  */
 	char *write_buf;
 	/* Size of write buffer.                */
-	int writebuf_size;
-	/* One of values in "enum ccs_proc_interface_index". */
-	u8 type;
+	size_t writebuf_size;
+	/* Type of interface. */
+	enum ccs_proc_interface_index type;
+	/* Users counter protected by ccs_io_buffer_list_lock. */
+	u8 users;
+	/* List for telling GC not to kfree() elements. */
+	struct list_head list;
 };
 
 /* Structure for /proc/ccs/profile interface. */
@@ -1404,159 +1514,69 @@ struct ccs_time {
 	u8 sec;
 };
 
+/* Structure for policy namespace. */
+struct ccs_policy_namespace {
+	/* Profile table. Memory is allocated as needed. */
+	struct ccs_profile *profile_ptr[CCS_MAX_PROFILES];
+	/* List of "struct ccs_group". */
+	struct list_head group_list[CCS_MAX_GROUP];
+	/* List of policy. */
+	struct list_head policy_list[CCS_MAX_POLICY];
+	/* The global ACL referred by "use_group" keyword. */
+	struct list_head acl_group[CCS_MAX_ACL_GROUPS];
+	/* List for connecting to ccs_namespace_list list. */
+	struct list_head namespace_list;
+	/* Profile version. Currently only 20100903 is defined. */
+	unsigned int profile_version;
+	/* Name of this namespace (e.g. "<kernel>", "</usr/sbin/httpd>" ). */
+	const char *name;
+};
+
 /* Prototype definition for "struct ccsecurity_operations". */
 
-void __init ccs_capability_init(void);
-void __init ccs_domain_init(void);
-void __init ccs_file_init(void);
+void __init ccs_permission_init(void);
 void __init ccs_mm_init(void);
-void __init ccs_mount_init(void);
-void __init ccs_network_init(void);
-void __init ccs_policy_io_init(void);
-void __init ccs_signal_init(void);
 
 /* Prototype definition for internal use. */
 
-bool ccs_address_matches_group(const bool is_ipv6, const u32 *address,
-			       const struct ccs_group *group);
-bool ccs_compare_number_union(const unsigned long value,
-			      const struct ccs_number_union *ptr);
-bool ccs_condition(struct ccs_request_info *r,
-		   const struct ccs_condition *cond);
-bool ccs_correct_domain(const unsigned char *domainname);
-bool ccs_correct_path(const char *filename);
-bool ccs_correct_word(const char *string);
-bool ccs_domain_def(const unsigned char *buffer);
-bool ccs_domain_quota_ok(struct ccs_request_info *r);
 bool ccs_dump_page(struct linux_binprm *bprm, unsigned long pos,
 		   struct ccs_page_dump *dump);
 bool ccs_memory_ok(const void *ptr, const unsigned int size);
-bool ccs_number_matches_group(const unsigned long min, const unsigned long max,
-			      const struct ccs_group *group);
-bool ccs_parse_name_union(const char *filename, struct ccs_name_union *ptr);
-bool ccs_parse_number_union(char *data, struct ccs_number_union *num);
-bool ccs_path_matches_pattern(const struct ccs_path_info *filename,
-			      const struct ccs_path_info *pattern);
-bool ccs_permstr(const char *string, const char *keyword);
-bool ccs_str_starts(char **src, const char *find);
-bool ccs_tokenize(char *buffer, char *w[], size_t size);
 char *ccs_encode(const char *str);
 char *ccs_encode2(const char *str, int str_len);
-char *ccs_init_log(struct ccs_request_info *r, int len, const char *fmt,
-		   va_list args);
-char *ccs_read_token(struct ccs_acl_param *param);
-char *ccs_realpath_from_path(struct path *path);
-const char *ccs_yesno(const unsigned int value);
+char *ccs_realpath(struct path *path);
 const char *ccs_get_exe(void);
-const struct ccs_path_info *ccs_compare_name_union
-(const struct ccs_path_info *name, const struct ccs_name_union *ptr);
-const struct ccs_path_info *ccs_get_domainname(struct ccs_acl_param *param);
 const struct ccs_path_info *ccs_get_name(const char *name);
-const struct ccs_path_info *ccs_path_matches_group
-(const struct ccs_path_info *pathname, const struct ccs_group *group);
-const struct in6_addr *ccs_get_ipv6_address(const struct in6_addr *addr);
-int ccs_close_control(struct file *file);
-int ccs_delete_domain(char *data);
-int ccs_env_perm(struct ccs_request_info *r, const char *env);
-int ccs_get_path(const char *pathname, struct path *path);
+int ccs_audit_log(struct ccs_request_info *r);
+int ccs_check_acl(struct ccs_request_info *r);
 int ccs_init_request_info(struct ccs_request_info *r, const u8 index);
-int ccs_lock(void);
-int ccs_open_control(const u8 type, struct file *file);
-int ccs_parse_ip_address(char *address, u16 *min, u16 *max);
-int ccs_path_permission(struct ccs_request_info *r, u8 operation,
-			const struct ccs_path_info *filename);
-int ccs_poll_control(struct file *file, poll_table *wait);
-int ccs_poll_log(struct file *file, poll_table *wait);
-int ccs_read_control(struct file *file, char __user *buffer,
-		     const int buffer_len);
-int ccs_supervisor(struct ccs_request_info *r, const char *fmt, ...)
-	__attribute__ ((format(printf, 2, 3)));
-int ccs_symlink_path(const char *pathname, struct ccs_path_info *name);
-int ccs_update_domain(struct ccs_acl_info *new_entry, const int size,
-		      struct ccs_acl_param *param,
-		      bool (*check_duplicate) (const struct ccs_acl_info *,
-					       const struct ccs_acl_info *),
-		      bool (*merge_duplicate) (struct ccs_acl_info *,
-					       struct ccs_acl_info *,
-					       const bool));
-int ccs_update_policy(struct ccs_acl_head *new_entry, const int size,
-		      bool is_delete, struct list_head *list,
-		      bool (*check_duplicate) (const struct ccs_acl_head *,
-					       const struct ccs_acl_head *));
-int ccs_write_aggregator(char *data, const bool is_delete);
-int ccs_write_capability(struct ccs_acl_param *param);
-int ccs_write_control(struct file *file, const char __user *buffer,
-		      const int buffer_len);
-int ccs_write_file(struct ccs_acl_param *param);
-int ccs_write_group(char *data, const bool is_delete, const u8 type);
-int ccs_write_inet_network(struct ccs_acl_param *param);
-int ccs_write_ipc(struct ccs_acl_param *param);
-int ccs_write_misc(struct ccs_acl_param *param);
-int ccs_write_reserved_port(char *data, const bool is_delete);
-int ccs_write_transition_control(char *data, const bool is_delete,
-				 const u8 type);
-int ccs_write_unix_network(struct ccs_acl_param *param);
-size_t ccs_del_condition(struct list_head *element);
-struct ccs_condition *ccs_get_condition(char *condition);
 struct ccs_domain_info *ccs_assign_domain(const char *domainname,
-					  const u8 profile, const u8 group,
 					  const bool transit);
-struct ccs_domain_info *ccs_find_domain(const char *domainname);
-struct ccs_group *ccs_get_group(const char *group_name, const u8 idx);
-struct ccs_profile *ccs_profile(const u8 profile);
 u8 ccs_get_config(const u8 profile, const u8 index);
-u8 ccs_parse_ulong(unsigned long *result, char **str);
 void *ccs_commit_ok(void *data, const unsigned int size);
-void ccs_check_acl(struct ccs_request_info *r,
-		   bool (*check_entry) (struct ccs_request_info *,
-					const struct ccs_acl_info *));
-void ccs_convert_time(time_t time, struct ccs_time *p);
+void ccs_del_acl(struct list_head *element);
+void ccs_del_condition(struct list_head *element);
 void ccs_fill_path_info(struct ccs_path_info *ptr);
 void ccs_get_attributes(struct ccs_obj_info *obj);
-void ccs_memory_free(const void *ptr, size_t size);
-void ccs_normalize_line(unsigned char *buffer);
-void ccs_print_ipv4(char *buffer, const int buffer_len, const u32 min_ip,
-		    const u32 max_ip);
-void ccs_print_ipv6(char *buffer, const int buffer_len,
-		    const struct in6_addr *min_ip,
-		    const struct in6_addr *max_ip);
-void ccs_print_ulong(char *buffer, const int buffer_len,
-		     const unsigned long value, const u8 type);
-void ccs_put_name_union(struct ccs_name_union *ptr);
-void ccs_put_number_union(struct ccs_number_union *ptr);
-void ccs_read_log(struct ccs_io_buffer *head);
-void ccs_run_gc(void);
+void ccs_notify_gc(struct ccs_io_buffer *head, const bool is_register);
 void ccs_transition_failed(const char *domainname);
-void ccs_unlock(const int idx);
-void ccs_update_stat(const u8 index);
 void ccs_warn_oom(const char *function);
 void ccs_write_log(struct ccs_request_info *r, const char *fmt, ...)
-	__attribute__ ((format(printf, 2, 3)));
-void ccs_write_log2(struct ccs_request_info *r, int len, const char *fmt,
-		    va_list args);
+	__printf(2, 3);
 
 /* Variable definition for internal use. */
 
 extern bool ccs_policy_loaded;
-extern const char * const ccs_condition_keyword[CCS_MAX_CONDITION_KEYWORD];
 extern const char * const ccs_dif[CCS_MAX_DOMAIN_INFO_FLAGS];
-extern const char * const ccs_mac_keywords[CCS_MAX_MAC_INDEX + CCS_MAX_MAC_CATEGORY_INDEX];
-extern const char * const ccs_mode[CCS_CONFIG_MAX_MODE];
-extern const char * const ccs_path_keyword[CCS_MAX_PATH_OPERATION];
-extern const char * const ccs_proto_keyword[CCS_SOCK_MAX];
-extern const char * const ccs_socket_keyword[CCS_MAX_NETWORK_OPERATION];
 extern const u8 ccs_c2mac[CCS_MAX_CAPABILITY_INDEX];
-extern const u8 ccs_index2category[CCS_MAX_MAC_INDEX];
 extern const u8 ccs_pn2mac[CCS_MAX_PATH_NUMBER_OPERATION];
 extern const u8 ccs_pnnn2mac[CCS_MAX_MKDEV_OPERATION];
 extern const u8 ccs_pp2mac[CCS_MAX_PATH2_OPERATION];
-extern struct ccs_domain_info ccs_acl_group[CCS_MAX_ACL_GROUPS];
 extern struct ccs_domain_info ccs_kernel_domain;
+extern struct list_head ccs_condition_list;
 extern struct list_head ccs_domain_list;
-extern struct list_head ccs_group_list[CCS_MAX_GROUP];
 extern struct list_head ccs_name_list[CCS_MAX_HASH];
-extern struct list_head ccs_policy_list[CCS_MAX_POLICY];
-extern struct list_head ccs_shared_list[CCS_MAX_LIST];
+extern struct list_head ccs_namespace_list;
 extern struct mutex ccs_policy_lock;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
 extern struct srcu_struct ccs_ss;
@@ -1580,39 +1600,6 @@ static inline bool ccs_pathcmp(const struct ccs_path_info *a,
 	return a->hash != b->hash || strcmp(a->name, b->name);
 }
 
-/**
- * ccs_same_name_union - Check for duplicated "struct ccs_name_union" entry.
- *
- * @a: Pointer to "struct ccs_name_union".
- * @b: Pointer to "struct ccs_name_union".
- *
- * Returns true if @a == @b, false otherwise.
- */
-static inline bool ccs_same_name_union(const struct ccs_name_union *a,
-				       const struct ccs_name_union *b)
-{
-	return a->filename == b->filename && a->group == b->group &&
-		a->is_group == b->is_group;
-}
-
-/**
- * ccs_same_number_union - Check for duplicated "struct ccs_number_union" entry.
- *
- * @a: Pointer to "struct ccs_number_union".
- * @b: Pointer to "struct ccs_number_union".
- *
- * Returns true if @a == @b, false otherwise.
- */
-static inline bool ccs_same_number_union(const struct ccs_number_union *a,
-					 const struct ccs_number_union *b)
-{
-	return a->values[0] == b->values[0] && a->values[1] == b->values[1]
-		&& a->group == b->group &&
-		a->value_type[0] == b->value_type[0] &&
-		a->value_type[1] == b->value_type[1] &&
-		a->is_group == b->is_group;
-}
-
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
 
 /**
@@ -1639,6 +1626,9 @@ static inline void ccs_read_unlock(const int idx)
 
 #else
 
+int ccs_lock(void);
+void ccs_unlock(const int idx);
+
 /**
  * ccs_read_lock - Take lock for protecting policy.
  *
@@ -1723,7 +1713,7 @@ static inline pid_t ccs_sys_getppid(void)
 {
 	pid_t pid;
 	rcu_read_lock();
-	pid = task_tgid_vnr(current->real_parent);
+	pid = task_tgid_vnr(rcu_dereference(current->real_parent));
 	rcu_read_unlock();
 	return pid;
 }
@@ -1890,20 +1880,6 @@ static inline void ccs_put_group(struct ccs_group *group)
 		atomic_dec(&group->head.users);
 }
 
-/**
- * ccs_put_ipv6_address - Drop reference on "struct ccs_ipv6addr".
- *
- * @addr: Pointer to "struct in6_addr". Maybe NULL.
- *
- * Returns nothing.
- */
-static inline void ccs_put_ipv6_address(const struct in6_addr *addr)
-{
-	if (addr)
-		atomic_dec(&container_of(addr, struct ccs_ipv6addr,
-					 addr)->head.users);
-}
-
 /**
  * ccs_put_name - Drop reference on "struct ccs_name".
  *
@@ -1938,8 +1914,10 @@ extern struct ccsecurity_exports ccsecurity_exports;
  */
 struct ccs_security {
 	struct list_head list;
-	const struct task_struct *task; /* Maybe NULL. */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 29)
+	const struct task_struct *task;
+#else
+	struct pid *pid; /* Maybe NULL. */
 	const struct cred *cred; /* Maybe NULL. */
 #endif
 	struct ccs_domain_info *ccs_domain_info;
@@ -1953,7 +1931,6 @@ bool ccs_used_by_cred(const struct ccs_domain_info *domain);
 int ccs_start_execve(struct linux_binprm *bprm, struct ccs_execve **eep);
 void ccs_finish_execve(int retval, struct ccs_execve *ee);
 void ccs_load_policy(const char *filename);
-extern spinlock_t ccs_task_security_list_lock;
 
 #define CCS_TASK_SECURITY_HASH_BITS 12
 #define CCS_MAX_TASK_SECURITY_HASH (1u << CCS_TASK_SECURITY_HASH_BITS)
@@ -2075,8 +2052,8 @@ static inline struct ccs_domain_info *ccs_task_domain(struct task_struct *task)
  * Returns pointer to "struct ccs_domain_info" for current thread.
  *
  * If current thread does not belong to a domain (which is true for initial
- * init_task in order to hide ccs_kernel_domain from this module), current
- * thread enters into ccs_kernel_domain.
+ * init_task in order to hide ccs_kernel_domain from this module),
+ * current thread enters into ccs_kernel_domain.
  */
 static inline struct ccs_domain_info *ccs_current_domain(void)
 {
@@ -2110,4 +2087,14 @@ static inline u32 ccs_current_flags(void)
 
 #endif
 
+/**
+ * ccs_current_namespace - Get "struct ccs_policy_namespace" for current thread.
+ *
+ * Returns pointer to "struct ccs_policy_namespace" for current thread.
+ */
+static inline struct ccs_policy_namespace *ccs_current_namespace(void)
+{
+	return ccs_current_domain()->ns;
+}
+
 #endif
