@@ -1,7 +1,7 @@
 /*
  * lsm.c
  *
- * Copyright (C) 2010-2011  Tetsuo Handa <penguin-kernel@I-love.SAKURA.ne.jp>
+ * Copyright (C) 2010-2012  Tetsuo Handa <penguin-kernel@I-love.SAKURA.ne.jp>
  *
  * Version: 1.0.24   2011/12/13
  */
@@ -649,6 +649,27 @@ static int ccs_path_chown(struct path *path, uid_t user, gid_t group)
 	return original_security_ops.path_chown(path, user, group);
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 3, 0)
+
+/**
+ * ccs_path_chmod - Check permission for chmod().
+ *
+ * @path: Pointer to "struct path".
+ * @mode: Mode.
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_path_chmod(struct path *path, umode_t mode)
+{
+	int rc = ccs_chmod_permission(path->dentry, path->mnt, mode);
+	if (rc)
+		return rc;
+	while (!original_security_ops.path_chmod);
+	return original_security_ops.path_chmod(path, mode);
+}
+
+#else
+
 /**
  * ccs_path_chmod - Check permission for chmod().
  *
@@ -668,6 +689,8 @@ static int ccs_path_chmod(struct dentry *dentry, struct vfsmount *vfsmnt,
 	return original_security_ops.path_chmod(dentry, vfsmnt, mode);
 }
 
+#endif
+
 /**
  * ccs_path_chroot - Check permission for chroot().
  *
@@ -814,6 +837,49 @@ static int ccs_inode_getattr(struct vfsmount *mnt, struct dentry *dentry)
 
 #if defined(CONFIG_SECURITY_PATH)
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 3, 0)
+
+/**
+ * ccs_path_mknod - Check permission for mknod().
+ *
+ * @dir:    Pointer to "struct path".
+ * @dentry: Pointer to "struct dentry".
+ * @mode:   Create mode.
+ * @dev:    Device major/minor number.
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_path_mknod(struct path *dir, struct dentry *dentry,
+			  umode_t mode, unsigned int dev)
+{
+	int rc = ccs_mknod_permission(dentry, dir->mnt, mode, dev);
+	if (rc)
+		return rc;
+	while (!original_security_ops.path_mknod);
+	return original_security_ops.path_mknod(dir, dentry, mode, dev);
+}
+
+/**
+ * ccs_path_mkdir - Check permission for mkdir().
+ *
+ * @dir:    Pointer to "struct path".
+ * @dentry: Pointer to "struct dentry".
+ * @mode:   Create mode.
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_path_mkdir(struct path *dir, struct dentry *dentry,
+			  umode_t mode)
+{
+	int rc = ccs_mkdir_permission(dentry, dir->mnt, mode);
+	if (rc)
+		return rc;
+	while (!original_security_ops.path_mkdir);
+	return original_security_ops.path_mkdir(dir, dentry, mode);
+}
+
+#else
+
 /**
  * ccs_path_mknod - Check permission for mknod().
  *
@@ -852,6 +918,8 @@ static int ccs_path_mkdir(struct path *dir, struct dentry *dentry, int mode)
 	return original_security_ops.path_mkdir(dir, dentry, mode);
 }
 
+#endif
+
 /**
  * ccs_path_rmdir - Check permission for rmdir().
  *
@@ -1118,6 +1186,49 @@ static int ccs_inode_create(struct inode *dir, struct dentry *dentry,
 
 #else
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 3, 0)
+
+/**
+ * ccs_inode_mknod - Check permission for mknod().
+ *
+ * @dir:    Pointer to "struct inode".
+ * @dentry: Pointer to "struct dentry".
+ * @mode:   Create mode.
+ * @dev:    Device major/minor number.
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_inode_mknod(struct inode *dir, struct dentry *dentry,
+			   umode_t mode, dev_t dev)
+{
+	int rc = ccs_mknod_permission(dentry, NULL, mode, dev);
+	if (rc)
+		return rc;
+	while (!original_security_ops.inode_mknod);
+	return original_security_ops.inode_mknod(dir, dentry, mode, dev);
+}
+
+/**
+ * ccs_inode_mkdir - Check permission for mkdir().
+ *
+ * @dir:    Pointer to "struct inode".
+ * @dentry: Pointer to "struct dentry".
+ * @mode:   Create mode.
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_inode_mkdir(struct inode *dir, struct dentry *dentry,
+			   umode_t mode)
+{
+	int rc = ccs_mkdir_permission(dentry, NULL, mode);
+	if (rc)
+		return rc;
+	while (!original_security_ops.inode_mkdir);
+	return original_security_ops.inode_mkdir(dir, dentry, mode);
+}
+
+#else
+
 /**
  * ccs_inode_mknod - Check permission for mknod().
  *
@@ -1156,6 +1267,8 @@ static int ccs_inode_mkdir(struct inode *dir, struct dentry *dentry, int mode)
 	return original_security_ops.inode_mkdir(dir, dentry, mode);
 }
 
+#endif
+
 /**
  * ccs_inode_rmdir - Check permission for rmdir().
  *
@@ -1249,6 +1362,29 @@ static int ccs_inode_link(struct dentry *old_dentry, struct inode *dir,
 	return original_security_ops.inode_link(old_dentry, dir, new_dentry);
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 3, 0)
+
+/**
+ * ccs_inode_create - Check permission for creat().
+ *
+ * @dir:    Pointer to "struct inode".
+ * @dentry: Pointer to "struct dentry".
+ * @mode:   Create mode.
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_inode_create(struct inode *dir, struct dentry *dentry,
+			    umode_t mode)
+{
+	int rc = ccs_mknod_permission(dentry, NULL, mode, 0);
+	if (rc)
+		return rc;
+	while (!original_security_ops.inode_create);
+	return original_security_ops.inode_create(dir, dentry, mode);
+}
+
+#else
+
 /**
  * ccs_inode_create - Check permission for creat().
  *
@@ -1270,6 +1406,8 @@ static int ccs_inode_create(struct inode *dir, struct dentry *dentry,
 
 #endif
 
+#endif
+
 #ifdef CONFIG_SECURITY_NETWORK
 
 #include <net/sock.h>
